"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.AdminClient = void 0;
const hatchet_error_1 = __importDefault(require("../../util/errors/hatchet-error"));
const workflow_run_ref_1 = __importDefault(require("../../util/workflow-run-ref"));
const grpc_helpers_1 = require("../../util/grpc-helpers");
const child_listener_client_1 = require("../../clients/listeners/run-listener/child-listener-client");
const workflows_1 = require("../../protoc/workflows");
class AdminClient {
    constructor(config, api, runs) {
        this.config = config;
        this.logger = config.logger(`Admin`, config.log_level);
        const { client, channel, factory } = (0, grpc_helpers_1.createGrpcClient)(config, workflows_1.WorkflowServiceDefinition);
        this.grpc = client;
        this.listenerClient = new child_listener_client_1.RunListenerClient(config, channel, factory, api);
        this.runs = runs;
    }
    /**
     * Run a new instance of a workflow with the given input. This will create a new workflow run and return the ID of the
     * new run.
     * @param workflowName the name of the workflow to run
     * @param input an object containing the input to the workflow
     * @param options an object containing the options to run the workflow
     * @returns the ID of the new workflow run
     */
    runWorkflow(workflowName, input, options) {
        return __awaiter(this, void 0, void 0, function* () {
            let computedName = workflowName;
            try {
                if (this.config.namespace && !workflowName.startsWith(this.config.namespace)) {
                    computedName = this.config.namespace + workflowName;
                }
                const inputStr = JSON.stringify(input);
                const resp = yield this.grpc.triggerWorkflow(Object.assign(Object.assign({ name: computedName, input: inputStr }, options), { additionalMetadata: (options === null || options === void 0 ? void 0 : options.additionalMetadata)
                        ? JSON.stringify(options === null || options === void 0 ? void 0 : options.additionalMetadata)
                        : undefined, priority: options === null || options === void 0 ? void 0 : options.priority }));
                const id = resp.workflowRunId;
                const ref = new workflow_run_ref_1.default(id, this.listenerClient, this.runs, options === null || options === void 0 ? void 0 : options.parentId);
                yield ref.getWorkflowRunId();
                return ref;
            }
            catch (e) {
                throw new hatchet_error_1.default(e.message);
            }
        });
    }
    /**
     * Run multiple workflows runs with the given input and options. This will create new workflow runs and return their IDs.
     * Order is preserved in the response.
     * @param workflowRuns an array of objects containing the workflow name, input, and options for each workflow run
     * @returns an array of workflow run references
     */
    runWorkflows(workflowRuns_1) {
        return __awaiter(this, arguments, void 0, function* (workflowRuns, batchSize = 500) {
            // Prepare workflows to be triggered in bulk
            const workflowRequests = workflowRuns.map(({ workflowName, input, options }) => {
                let computedName = workflowName;
                if (this.config.namespace && !workflowName.startsWith(this.config.namespace)) {
                    computedName = this.config.namespace + workflowName;
                }
                const inputStr = JSON.stringify(input);
                return Object.assign(Object.assign({ name: computedName, input: inputStr }, options), { additionalMetadata: (options === null || options === void 0 ? void 0 : options.additionalMetadata)
                        ? JSON.stringify(options.additionalMetadata)
                        : undefined });
            });
            const batches = this.batchWorkflows(workflowRequests, batchSize);
            this.logger.debug(`batching ${batches.length} batches`);
            try {
                const allResults = [];
                // Process each batch
                for (let batchIndex = 0; batchIndex < batches.length; batchIndex += 1) {
                    const batch = batches[batchIndex];
                    const { requests, originalIndices } = batch;
                    // Call the bulk trigger workflow method for this batch
                    const bulkTriggerWorkflowResponse = yield this.grpc.bulkTriggerWorkflow(workflows_1.BulkTriggerWorkflowRequest.create({
                        workflows: requests,
                    }));
                    this.logger.debug(`batch ${batchIndex + 1} of ${batches.length}`);
                    // Map the results back to their original indices
                    const batchResults = bulkTriggerWorkflowResponse.workflowRunIds.map((resp, index) => {
                        const originalIndex = originalIndices[index];
                        const { options } = workflowRuns[originalIndex];
                        return new workflow_run_ref_1.default(resp, this.listenerClient, this.runs, options === null || options === void 0 ? void 0 : options.parentId);
                    });
                    allResults.push(...batchResults);
                }
                return allResults;
            }
            catch (e) {
                throw new hatchet_error_1.default(e.message);
            }
        });
    }
    batchWorkflows(workflowRequests, batchSize) {
        const payloadLimit = 4 * 1024 * 1024; // 4MB limit
        const batches = [];
        let currentBatch = [];
        let currentBatchIndices = [];
        let currentBatchSize = 0;
        for (let i = 0; i < workflowRequests.length; i += 1) {
            const request = workflowRequests[i];
            const requestSize = Buffer.byteLength(JSON.stringify(request), 'utf8');
            // Check if adding this request would exceed either the payload limit or batch size
            if (currentBatchSize + requestSize > payloadLimit || currentBatch.length >= batchSize) {
                // If we have a batch, add it to batches
                if (currentBatch.length > 0) {
                    batches.push({
                        requests: currentBatch,
                        originalIndices: currentBatchIndices,
                    });
                }
                // Start a new batch
                currentBatch = [request];
                currentBatchIndices = [i];
                currentBatchSize = requestSize;
            }
            else {
                // Add to current batch
                currentBatch.push(request);
                currentBatchIndices.push(i);
                currentBatchSize += requestSize;
            }
        }
        // Add the last batch if it exists
        if (currentBatch.length > 0) {
            batches.push({
                requests: currentBatch,
                originalIndices: currentBatchIndices,
            });
        }
        return batches;
    }
    putRateLimit(key, limit, duration) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.grpc.putRateLimit({
                key,
                limit,
                duration,
            });
        });
    }
}
exports.AdminClient = AdminClient;
